# -*- coding: utf-8 -*-
"""
Copyright (C) 2009-2017 All Rights Reserved. Luiz Fernando Maciel França
              www.siginformatica.com.br <luiz@siginformatica.com.br>
                                -------

Copyleft (CC) 2017, Creative Commons
Autor - Pierre Geraldo Moura Barbosa <barbosa.pierre[at]gmail.com>
================================================================================
Atribuicao-Nao Comercial-Compartilhada Igual 3.0 Nao Adaptada (CC BY-NC-SA 3.0)
http://creativecommons.org/licenses/by-nc-sa/3.0/deed.pt_BR
================================================================================
"""
from nfe.nfe_400.manual_600.cancnfe_evento import EnvEvento, RetEnvEvento, ProcEventoNFe
from nfe.nfe_400.manual_600.inutnfe import InutNFe, RetInutNFe, ProcInutNFe
from nfe.nfe_400.manual_600.consrecinfe import ConsReciNFe, RetConsReciNFe
from nfe.pysped.nfe.httplib_extras import HTTPSSignedConnection
from nfe.nfe_400.processador import ProcessoNFe, ProcessadorNFe
from nfe.nfe_400.manual_600.envinfe import EnviNFe, RetEnviNFe
from nfe.nfe_400.manual_600.conssitnfe import ConsSitNFe, RetConsSitNFe
from nfe.pysped.xml_sped.certificado import Certificado
from nfe.pysped.nfe.webservices_flags import UF_CODIGO
from nfe.nfce.webservices_flag import *
from lxml.etree import XMLSyntaxError
from datetime import datetime
from ssl import SSLError
import webservices
import tempfile
import logging
import pytz
import time
import os


class ProcessadorNFCe(ProcessadorNFe):
    """
    Processador da nota fiscal eletronica
    :author: Pierre Barbosa <pierre.barbosa[at]siginformatica.com.br>
    """
    def __init__(self, certify=None, key=None, environment=2, state=u'SVRS', contingency=u''):
        super(ProcessadorNFCe, self).__init__(certify, key, environment, state, contingency)
        self.estado = state  # Inicialmente os estados utilizam o ambiente do Rio Grande do Sul
        self.certificado = Certificado()
        self.certificado.certificado = certify
        self.certificado.chave = key
        self.certificado.cert_str = certify
        self.certificado.key_str = key
        self.ws = webservices.ESTADO_WS[self.versao][self.estado or 'SVRS']
        self.tentativas = 5
        self.logger = logging.getLogger('SEFAZ-NFCe')  # Registro de log

    @staticmethod
    def remove_accents(text):
        from unicodedata import normalize, category
        text = text.encode('utf-8') if type(text) == unicode else text
        text = filter(lambda c: category(c) != 'Mn', normalize('NFKD', text.decode('utf-8')))
        return text.encode('ascii', 'replace')

    def __init_soap_envio(self, servico, envio):
        self._soap_envio.webservice = self.ws[self.ambiente][servico]['webservice']
        self._soap_envio.metodo = self.ws[self.ambiente][servico]['metodo']
        self._soap_envio.cUF = UF_CODIGO[self.estado]
        self._soap_envio.namespace = self.ws[self.ambiente][servico]['namespace']
        self._soap_envio.envio = envio

    def __init_soap_retorno(self, servico, resposta):
        self._soap_retorno.webservice = self.ws[self.ambiente][servico]['webservice']
        self._soap_retorno.metodo = self.ws[self.ambiente][servico]['metodo']
        self._soap_retorno.resposta = resposta

    def __check_cert_validate(self):
        today = datetime.now(pytz.timezone('UTC')).astimezone(pytz.timezone('Brazil/East'))
        validate = self.certificado.data_fim_validade
        if validate.strftime('%Y%m%d') < today.strftime('%Y%m%d'):
            return False

        return True

    def __init_certify(self, cert, key):
        self.certificado.cert_str = self.certificado.certificado.encode()
        self.certificado.key_str = self.certificado.chave.encode()
        self.certificado.prepara_certificado_arquivo_pfx()

    def _conectar_servico(self, servico, envio, resposta):
        self.logger.info(u'>> %s-%s v-%s' % (
            self.ambiente, u'Producao' if self.ambiente == 1 else u'Homologacao', self.versao))
        self.logger.info(self.contingencia or u'>> %s/%s' % (self.estado, UF_CODIGO[self.estado]))
        map(lambda m: self.logger.info(u'>> %s: %s' % m), self.ws[self.ambiente][servico].items())
        self.__init_soap_envio(servico, envio)
        self.__init_soap_retorno(servico, resposta)
        _servidor = self.ws[self.ambiente][servico]['servidor']
        _url = self.ws[self.ambiente][servico]['url']

        # Verifica vencimento do certificado
        cert, key = self.certificado.certificado.encode(), self.certificado.chave.encode()
        self.__init_certify(cert, key)
        if not self.__check_cert_validate():
            self._soap_retorno.resposta.cStat.valor = u'999'
            self._soap_retorno.resposta.xMotivo.valor = u'Certificado digital com data expirada.'
            self.logger.critical(self._soap_retorno.resposta.xMotivo.valor)

        # Arquivos de chave e certificado digital que serao utilizados na conexao https
        key_file = tempfile.NamedTemporaryFile('w')
        key_file.write(self.certificado.key_str)
        key_file.flush()
        cert_file = tempfile.NamedTemporaryFile('w')
        cert_file.write(self.certificado.cert_str)
        cert_file.flush()
        ca_file = open(os.path.dirname(
            os.path.abspath(__file__)) + '/../pysped/xml_sped/cadeia-certificadora/cadeia.pem')

        con = HTTPSSignedConnection(_servidor, key_file=key_file.name, cert_file=cert_file.name,
                                    ca_file=ca_file.name)
        request_url = u'' + os.path.join('/', _url)
        request_body = self._soap_envio.xml.encode()
        request_header = self._soap_envio.header
        try:
            con.request('POST', request_url, body=request_body, headers=request_header)
        except Exception as error:
            self.logger.error(u'<<< erro conexao: %s >>>' % error)
            self._soap_retorno.resposta.cStat.valor = u'999'
            self._soap_retorno.resposta.xMotivo.valor = error
        else:
            for tentativa in range(self.tentativas):
                self.logger.error(u'<<< Tentativa [%d] >>>' % (tentativa + 1))
                try:
                    # Tentativas de conexao com tempos subsequentes maiores para evitar retorno com
                    # falha
                    time.sleep(tentativa * 2)  # Tempo subsequente maior para a proxima tentativa
                    resp = con.getresponse()
                    self._soap_retorno.resposta.version = resp.version
                    self._soap_retorno.resposta.status = resp.status
                    self._soap_retorno.resposta.msg = resp.msg
                    reason = resp.reason.decode('utf-8')
                    result = resp.read().decode('utf-8')
                    self._soap_retorno.resposta.reason = reason
                    self._soap_retorno.resposta.original = result
                    self._soap_retorno.xml = result
                except SSLError as error:
                    self._soap_retorno.resposta.cStat.valor = u'999'
                    self._soap_retorno.resposta.xMotivo.valor = str(error)
                    self.logger.error(u'<<< erro conexao: %s >>>' % str(error))
                    continue
                except XMLSyntaxError as error:  # Erro de estrutura do XML de retorno
                    self._soap_retorno.resposta.cStat.valor = u'999'
                    self._soap_retorno.resposta.xMotivo.valor = error.msg
                    self.logger.error(u'<<< erro conexao: 999-%s' % error.msg)
                    break
                except Exception as error:
                    self._soap_retorno.resposta.cStat.valor = u'999'
                    self._soap_retorno.resposta.xMotivo.valor = error.args[1]
                    self.logger.error(u'<<< erro conexao: %s-%s' % error.args)
                    break
                else:
                    self.logger.info(u'>> URL do servico: %s' % os.path.join(_servidor, _url))
                    self.logger.info(u'>> Metodo da requisição: POST')
                    self.logger.info(u'>> URL da requisição: %s' % request_url)
                    self.logger.info(u'>> Cabecalho da requisicao: %s' % request_header)
                    self.logger.info(u'>> XML envio: %s' % request_body)
                    self.logger.info(u'<< Status: %s-%s' % (resp.status, reason))
                    self.logger.info(u'<< XML resposta: %s' % result)
                    self.logger.info(u'<< Mensagem servidor: %s' % resp.msg)
                    con.close()  # Fecha a conexao
                    break  # Termina as tentativas, jah que obteve sucesso
        finally:
            # Remove os arquivos temporarios: chave e certificado
            key_file.close()
            cert_file.close()

    def processar_lote(self, lote):
        self.logger.info(u'*** montagem do lote de NFC-e ***')

        # Assina e valida todas as NFC-e antes da transmissao
        lote = lote if isinstance(lote, list) else [lote]  # Sempre uma lista!
        for nfce in lote:
            nfce.monta_chave()
            nfce.xml = self.remove_accents(nfce.xml)
            self.certificado.assina_xmlnfe(nfce)
            nfce.validar()
            self.logger.info(u'%s=> %s' % (lote.index(nfce) + 1, nfce.infNFe.Id.valor))

        envio = EnviNFe()
        resposta = RetEnviNFe()
        processo = ProcessoNFe(WS_AUTORIZACAO, envio, resposta)
        envio.NFe = lote
        envio.idLote.valor = str(self.ambiente) + datetime.now().strftime('%Y%m%d%H%M%S')
        envio.validar()
        self.logger.info(u'*** envio do lote de NFC-e ***')
        self._conectar_servico(WS_AUTORIZACAO, envio, resposta)
        if processo.resposta.cStat.valor == u'103':
            processo = self.consultar_recibo(processo.resposta.infRec.nRec.valor)

        return processo

    def consultar_recibo(self, recibo):
        envio = ConsReciNFe()
        resposta = RetConsReciNFe()
        processo = ProcessoNFe(WS_RET_AUTORIZACAO, envio, resposta)
        envio.tpAmb.valor = self.ambiente
        envio.nRec.valor = recibo
        envio.validar()
        self.logger.info(u'*** consulta de lote NFC-e ***')
        self.logger.info(u'Lote %s' % recibo)
        while True:  # Tentativas para servico em processamento
            self._conectar_servico(WS_RET_AUTORIZACAO, envio, resposta)
            if processo.resposta.cStat.valor == u'105':
                continue
            else:
                break

        return processo

    def consultar_nfce(self, chave):
        envio = ConsSitNFe()
        resposta = RetConsSitNFe()
        processo = ProcessoNFe(WS_CONSULTA_PROTOCOLO, envio, resposta)
        envio.tpAmb.valor = self.ambiente
        envio.chNFe.valor = chave
        envio.validar()
        self.logger.info(u'*** consulta NFC-e ***')
        self._conectar_servico(WS_CONSULTA_PROTOCOLO, envio, resposta)

        return processo

    def cancelar_nfce(self, chave, protocolo, justificativa):
        envio = EnvEvento()
        resposta = RetEnvEvento()
        processo = ProcessoNFe(WS_RECEPCAO_EVENTO, envio, resposta)
        envio.infEvento.tpAmb.valor = self.ambiente
        envio.infEvento.chNFe.valor = chave
        envio.infEvento.nProt.valor = protocolo
        envio.infEvento.xJust.valor = justificativa
        envio.infEvento.cOrgao.valor = UF_CODIGO[self.estado]
        dt = datetime.utcnow()
        dt = dt.replace(tzinfo=pytz.utc)
        data_hora_tz = datetime.astimezone(dt, pytz.timezone("Brazil/East"))
        data_evento = datetime.strftime(data_hora_tz, u'%Y-%m-%dT%H:%M:%S') + str(data_hora_tz)[-6:]
        envio.infEvento.dhEvento.valor = data_evento
        envio.infEvento.CNPJ.valor = chave[6:20]
        self.certificado.assina_xmlnfe(envio)
        envio.validar()
        self.logger.info(u'*** cancelamento de NFC-e ***')
        self._conectar_servico(WS_RECEPCAO_EVENTO, envio, resposta)
        if resposta.cStat.valor == u'128':
            processo_cancelamento_nfe = ProcEventoNFe()
            processo_cancelamento_nfe.envEvento = envio
            processo_cancelamento_nfe.retEnvEvento = resposta
            processo_cancelamento_nfe.validar()
            processo.processo_cancelamento_nfe = processo_cancelamento_nfe

        return processo

    def inutilizar_nfce(self, chave, justificativa):
        ano = chave[2:4]
        cnpj = chave[6:20]
        serie = chave[22:25]
        inicio = fim = chave[25:34]
        return self.inutilizar_faixa_numeracao(ano, cnpj, serie, inicio, fim, justificativa)

    def inutilizar_faixa_numeracao(self, ano, cnpj, serie, inicio=1, fim=1, justificativa=None):
        envio = InutNFe()
        resposta = RetInutNFe()
        processo = ProcessoNFe(WS_INUTILIZACAO, envio, resposta)
        envio.infInut.tpAmb.valor = self.ambiente
        envio.infInut.cUF.valor = UF_CODIGO[self.estado]
        envio.infInut.ano.valor = ano or datetime.now().strftime(u'%y')
        envio.infInut.CNPJ.valor = cnpj
        envio.infInut.serie.valor = int(serie)
        envio.infInut.nNFIni.valor = int(inicio)
        envio.infInut.nNFFin.valor = int(fim)
        envio.infInut.xJust.valor = justificativa
        envio.infInut.mod.valor = 65  # NFC-e
        envio.gera_nova_chave()
        self.certificado.prepara_certificado_arquivo_pfx()
        self.certificado.assina_xmlnfe(envio)
        envio.validar()
        self.logger.info(u'*** inutilizar faixa de numeracao NFC-e ***')
        self.logger.info(u'>> Faixa: %03d a %03d - Serie %s' % (int(inicio), int(fim), int(serie)))
        self.logger.info(u'>> ID%s (chave)' % envio.chave)
        self._conectar_servico(WS_INUTILIZACAO, envio, resposta)
        processo_inutilizacao_nfe = ProcInutNFe()

        # Se for autorizada, monta o processo de inutilização
        if resposta.infInut.cStat.valor == u'102':
            processo_inutilizacao_nfe.inutNFe = envio
            processo_inutilizacao_nfe.retInutNFe = resposta
            processo_inutilizacao_nfe.validar()
            processo.processo_inutilizacao_nfe = processo_inutilizacao_nfe

        return processo

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
